from allennlp_models.lm.bidirectional_lm import BidirectionalLanguageModel
from allennlp_models.lm.bidirectional_lm_token_embedder import (
    BidirectionalLanguageModelTokenEmbedder,
)
from allennlp_models.lm.bidirectional_lm_transformer import BidirectionalLanguageModelTransformer
from allennlp_models.lm.language_model import LanguageModel
from allennlp_models.lm.language_model_token_embedder import LanguageModelTokenEmbedder
from allennlp_models.lm.masked_language_model import MaskedLanguageModel
from allennlp_models.lm.masked_language_model_predictor import MaskedLanguageModelPredictor
from allennlp_models.lm.masked_language_model_reader import MaskedLanguageModelingReader
from allennlp_models.lm.next_token_lm import NextTokenLM
from allennlp_models.lm.next_token_lm_predictor import NextTokenLMPredictor
from allennlp_models.lm.next_token_lm_reader import NextTokenLMReader
from allennlp_models.lm.simple_language_modeling_reader import SimpleLanguageModelingDatasetReader



from allennlp_models.rc.bidaf import BidafEnsemble
from allennlp_models.rc.bidaf import BidirectionalAttentionFlow
from allennlp_models.rc.dialog_qa import DialogQAPredictor
from allennlp_models.rc.dialog_qa import DialogQA
from allennlp_models.rc.dialog_qa import QuACReader
from allennlp_models.rc.qanet import QaNet
from allennlp_models.rc.qanet import NumericallyAugmentedQaNet
from allennlp_models.rc.transformer_qa import TransformerQAPredictor
from allennlp_models.rc.transformer_qa import TransformerQA
from allennlp_models.rc.transformer_qa import TransformerSquadReader



"""
Coreference resolution is defined as follows: given a document, find and cluster entity mentions.
"""

from allennlp_models.coref.conll_reader import ConllCorefReader
from allennlp_models.coref.preco_reader import PrecoReader
from allennlp_models.coref.winobias_reader import WinobiasReader
from allennlp_models.coref.coref_model import CoreferenceResolver
from allennlp_models.coref.coref_predictor import CorefPredictor



from allennlp_models.rc.bidaf.bidaf_model import BidirectionalAttentionFlow
from allennlp_models.rc.bidaf.bidaf_ensemble import BidafEnsemble
from allennlp_models.rc.bidaf.bidaf_predictor import ReadingComprehensionPredictor



"""
This evaluation script relies heavily on the one for DROP (``allennlp/tools/drop_eval.py``). We need a separate
script for Quoref only because the data formats are slightly different.
"""

import json
from typing import Dict, Tuple, List, Any, Optional
import argparse
import numpy as np

from . import drop_eval


def _get_answers_from_data(annotations: Dict[str, Any]) -> Dict[str, List[str]]:
    """
    If the annotations file is in the same format as the original data files, this method can be used to extract a
    dict of query ids and answers.
    """
    answers_dict: Dict[str, List[str]] = {}
    for article_info in annotations["data"]:
        for paragraph_info in article_info["paragraphs"]:
            for qa_pair in paragraph_info["qas"]:
                query_id = qa_pair["id"]
                candidate_answers = [answer["text"] for answer in qa_pair["answers"]]
                answers_dict[query_id] = candidate_answers
    return answers_dict


def evaluate_json(
    annotations: Dict[str, Any], predicted_answers: Dict[str, Any]
) -> Tuple[float, float]:
    """
    Takes gold annotations and predicted answers and  evaluates the predictions for each question
    in the gold annotations.  Both JSON dictionaries must have query_id keys, which are used to
    match predictions to gold annotations.

    The ``predicted_answers`` JSON must be a dictionary keyed by query id, where the value is a
    list of strings (or just one string) that is the answer.
    The ``annotations`` are assumed to have either the format of the dev set in the Quoref data release, or the
    same format as the predicted answers file.
    """
    instance_exact_match = []
    instance_f1 = []
    if "data" in annotations:
        # We're looking at annotations in the original data format. Let's extract the answers.
        annotated_answers = _get_answers_from_data(annotations)
    else:
        annotated_answers = annotations
    for query_id, candidate_answers in annotated_answers.items():
        max_em_score = 0.0
        max_f1_score = 0.0
        if query_id in predicted_answers:
            predicted = predicted_answers[query_id]
            gold_answer = tuple(candidate_answers)
            em_score, f1_score = drop_eval.get_metrics(predicted, gold_answer)
            if gold_answer[0].strip() != "":
                max_em_score = max(max_em_score, em_score)
                max_f1_score = max(max_f1_score, f1_score)
        else:
            print("Missing prediction for question: {}".format(query_id))
            max_em_score = 0.0
            max_f1_score = 0.0
        instance_exact_match.append(max_em_score)
        instance_f1.append(max_f1_score)

    global_em = np.mean(instance_exact_match)
    global_f1 = np.mean(instance_f1)
    print("Exact-match accuracy {0:.2f}".format(global_em * 100))
    print("F1 score {0:.2f}".format(global_f1 * 100))
    print("{0:.2f}   &   {1:.2f}".format(global_em * 100, global_f1 * 100))
    return global_em, global_f1


def evaluate_prediction_file(
    prediction_path: str, gold_path: str, output_path: Optional[str] = None
) -> Tuple[float, float]:
    """
    Takes a prediction file and a gold file and evaluates the predictions for each question in the gold file.  Both
    files must be json formatted and must have query_id keys, which are used to match predictions to gold
    annotations. Writes a json with global_em and global_f1 metrics to file at the specified output
    path, unless None is passed as output path.
    """
    predicted_answers = json.load(open(prediction_path, encoding="utf-8"))
    annotations = json.load(open(gold_path, encoding="utf-8"))
    global_em, global_f1 = evaluate_json(annotations, predicted_answers)

    # Output predictions to file if an output path is given
    if output_path is not None:
        output_dict = {"global_em": global_em, "global_f1": global_f1}

        with open(output_path, "w", encoding="utf8") as outfile:
            json.dump(output_dict, outfile)

    return (global_em, global_f1)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Evaluate Quoref predictions")
    parser.add_argument(
        "--gold_path",
        type=str,
        required=False,
        default="quoref-test-v0.1.json",
        help="location of the gold file",
    )
    parser.add_argument(
        "--prediction_path",
        type=str,
        required=False,
        default="sample_predictions.json",
        help="location of the prediction file",
    )
    parser.add_argument(
        "--output_path",
        type=str,
        required=False,
        default=None,
        help="location of the output metrics file",
    )

    args = parser.parse_args()
    evaluate_prediction_file(args.prediction_path, args.gold_path, args.output_path)



from allennlp_models.syntax.srl.srl_eval_scorer import SrlEvalScorer
from allennlp_models.syntax.srl.srl_bert_model import SrlBert
from allennlp_models.syntax.srl.srl_model import SemanticRoleLabeler
from allennlp_models.syntax.srl.srl_predictor import SemanticRoleLabelerPredictor
from allennlp_models.syntax.srl.srl_reader import SrlReader
from allennlp_models.syntax.srl.openie_predictor import OpenIePredictor



from allennlp_models.nli.snli_reader import SnliReader
from allennlp_models.nli.decomposable_attention_model import DecomposableAttention
from allennlp_models.nli.decomposable_attention_predictor import DecomposableAttentionPredictor
from allennlp_models.nli.bimpm_model import BiMpm
from allennlp_models.nli.esim_model import ESIM
from allennlp_models.nli.quora_paraphrase_reader import QuoraParaphraseDatasetReader



from allennlp_models.syntax.biaffine_dependency_parser import BiaffineDependencyParser
from allennlp_models.syntax.biaffine_dependency_parser import BiaffineDependencyParserPredictor
from allennlp_models.syntax.biaffine_dependency_parser import UniversalDependenciesDatasetReader
from allennlp_models.syntax.constituency_parser import SpanConstituencyParser
from allennlp_models.syntax.constituency_parser import ConstituencyParserPredictor
from allennlp_models.syntax.constituency_parser import PennTreeBankConstituencySpanDatasetReader
from allennlp_models.syntax.srl import SrlBert
from allennlp_models.syntax.srl import SrlReader
from allennlp_models.syntax.srl import SemanticRoleLabelerPredictor
from allennlp_models.syntax.srl import SemanticRoleLabeler
from allennlp_models.syntax.srl import SrlEvalScorer
from allennlp_models.syntax.srl import OpenIePredictor
from allennlp_models.syntax.semantic_dependencies import GraphParser
from allennlp_models.syntax.semantic_dependencies import SemanticDependenciesDatasetReader



from allennlp_models.seq2seq.auto_regressive_seq_decoder import AutoRegressiveSeqDecoder
from allennlp_models.seq2seq.composed_seq2seq_model import ComposedSeq2Seq
from allennlp_models.seq2seq.copynet import CopyNetSeq2Seq
from allennlp_models.seq2seq.copynet import CopyNetDatasetReader
from allennlp_models.seq2seq.decoder_net import DecoderNet
from allennlp_models.seq2seq.lstm_cell_decoder_net import LstmCellDecoderNet
from allennlp_models.seq2seq.seq2seq_predictor import Seq2SeqPredictor
from allennlp_models.seq2seq.seq_decoder import SeqDecoder
from allennlp_models.seq2seq.simple_seq2seq_model import SimpleSeq2Seq
from allennlp_models.seq2seq.simple_seq2seq_predictor import SimpleSeq2SeqPredictor
from allennlp_models.seq2seq.stacked_self_attention_decoder_net import (
    StackedSelfAttentionDecoderNet,
)
from allennlp_models.seq2seq.seq2seq_reader import Seq2SeqDatasetReader



from allennlp_models.rc.qanet.drop_reader import DropReader
from allennlp_models.rc.qanet.naqanet_model import NumericallyAugmentedQaNet
from allennlp_models.rc.qanet.qanet_model import QaNet
from allennlp_models.rc.qanet.qanet_encoder import QaNetEncoder
from allennlp_models.rc.qanet.qanet_encoder import QaNetEncoderBlock



from allennlp.models import load_archive
from allennlp.predictors import Predictor, SentenceTaggerPredictor

from allennlp_models.coref import CorefPredictor
from allennlp_models.nli import DecomposableAttentionPredictor
from allennlp_models.rc.bidaf import ReadingComprehensionPredictor
from allennlp_models.syntax import (
    SemanticRoleLabelerPredictor,
    OpenIePredictor,
    ConstituencyParserPredictor,
    BiaffineDependencyParserPredictor,
)

# flake8: noqa: E501


def _load_predictor(archive_file: str, predictor_name: str) -> Predictor:
    """
    Helper to load the desired predictor from the given archive.
    """
    archive = load_archive(archive_file)
    return Predictor.from_archive(archive, predictor_name)


def bert_srl_shi_2019() -> SemanticRoleLabelerPredictor:
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/bert-base-srl-2020.03.24.tar.gz",
        "semantic-role-labeling",
    )
    return predictor


def bidirectional_attention_flow_seo_2017() -> ReadingComprehensionPredictor:
    """
    Reading Comprehension

    Based on `BiDAF (Seo et al, 2017) <https://www.semanticscholar.org/paper/Bidirectional-Attention-Flow-for-Machine-Comprehen-Seo-Kembhavi/007ab5528b3bd310a80d553cccad4b78dc496b02>`_
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/bidaf-model-2020.03.19.tar.gz",
        "reading-comprehension",
    )
    return predictor


def naqanet_dua_2019() -> ReadingComprehensionPredictor:
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/naqanet-2020.02.19.tar.gz",
        "reading-comprehension",
    )
    return predictor


def open_information_extraction_stanovsky_2018() -> OpenIePredictor:
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/openie-model.2020.03.26.tar.gz",
        "open-information-extraction",
    )
    return predictor


def decomposable_attention_with_elmo_parikh_2017() -> DecomposableAttentionPredictor:
    """
    Textual Entailment

    Based on `Parikh et al, 2017 <https://www.semanticscholar.org/paper/A-Decomposable-Attention-Model-for-Natural-Languag-Parikh-T%C3%A4ckstr%C3%B6m/07a9478e87a8304fc3267fa16e83e9f3bbd98b27>`_
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/decomposable-attention-elmo-2020.04.09.tar.gz",
        "textual-entailment",
    )
    return predictor


def neural_coreference_resolution() -> CorefPredictor:
    """
    Coreference Resolution
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/coref-spanbert-large-2020.02.27.tar.gz",
        "coreference-resolution",
    )
    return predictor


def named_entity_recognition_with_elmo_peters_2018() -> SentenceTaggerPredictor:
    """
    Named Entity Recognition

    Based on `Deep contextualized word representations <https://arxiv.org/abs/1802.05365>`_
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/ner-model-2020.02.10.tar.gz",
        "sentence-tagger",
    )
    return predictor


def fine_grained_named_entity_recognition_with_elmo_peters_2018() -> SentenceTaggerPredictor:
    """
    Fine Grained Named Entity Recognition
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/fine-grained-ner-model-elmo-2018.12.21.tar.gz",
        "sentence-tagger",
    )
    return predictor


def span_based_constituency_parsing_with_elmo_joshi_2018() -> ConstituencyParserPredictor:
    """
    Constituency Parsing

    Based on `Minimal Span Based Constituency Parser (Stern et al, 2017) <https://www.semanticscholar.org/paper/A-Minimal-Span-Based-Neural-Constituency-Parser-Stern-Andreas/593e4e749bd2dbcaf8dc25298d830b41d435e435>`_ but with ELMo embeddings
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/elmo-constituency-parser-2020.02.10.tar.gz",
        "constituency-parser",
    )
    return predictor


def biaffine_parser_universal_dependencies_todzat_2017() -> BiaffineDependencyParserPredictor:
    """
    Biaffine Dependency Parser (Stanford Dependencies)

    Based on `Dozat and Manning, 2017 <https://arxiv.org/pdf/1611.01734.pdf>`_
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/biaffine-dependency-parser-ptb-2020.04.06.tar.gz",
        "biaffine-dependency-parser",
    )
    return predictor


def esim_nli_with_elmo_chen_2017() -> DecomposableAttentionPredictor:
    """
    ESIM

    Based on `Enhanced LSTM for Natural Language Inference <https://arxiv.org/pdf/1609.06038.pdf>`_ and uses ELMo
    """
    predictor = _load_predictor(
        "https://storage.googleapis.com/allennlp-public-models/decomposable-attention-elmo-2020.04.09.tar.gz",
        "textual-entailment",
    )
    return predictor



from typing import Dict, List, TextIO, Optional, Any
import warnings

from overrides import overrides
import torch
from torch.nn.modules import Linear, Dropout
import torch.nn.functional as F

from allennlp.common.checks import check_dimensions_match
from allennlp.data import TextFieldTensors, Vocabulary
from allennlp.modules import Seq2SeqEncoder, TimeDistributed, TextFieldEmbedder
from allennlp.modules.token_embedders import Embedding
from allennlp.models.model import Model
from allennlp.nn import InitializerApplicator
from allennlp.nn.util import get_text_field_mask, sequence_cross_entropy_with_logits
from allennlp.nn.util import get_lengths_from_binary_sequence_mask, viterbi_decode
from .srl_eval_scorer import SrlEvalScorer, DEFAULT_SRL_EVAL_PATH
from .util import convert_bio_tags_to_conll_format, write_bio_formatted_tags_to_file


@Model.register("srl")
class SemanticRoleLabeler(Model):
    """
    This model performs semantic role labeling using BIO tags using Propbank semantic roles.
    Specifically, it is an implementation of [Deep Semantic Role Labeling - What works
    and what's next](https://www.aclweb.org/anthology/P17-1044).

    This implementation is effectively a series of stacked interleaved LSTMs with highway
    connections, applied to embedded sequences of words concatenated with a binary indicator
    containing whether or not a word is the verbal predicate to generate predictions for in
    the sentence. Additionally, during inference, Viterbi decoding is applied to constrain
    the predictions to contain valid BIO sequences.

    Specifically, the model expects and outputs IOB2-formatted tags, where the
    B- tag is used in the beginning of every chunk (i.e. all chunks start with the B- tag).

    # Parameters

    vocab : `Vocabulary`, required
        A Vocabulary, required in order to compute sizes for input/output projections.
    text_field_embedder : `TextFieldEmbedder`, required
        Used to embed the `tokens` `TextField` we get as input to the model.
    encoder : `Seq2SeqEncoder`
        The encoder (with its own internal stacking) that we will use in between embedding tokens
        and predicting output tags.
    binary_feature_dim : int, required.
        The dimensionality of the embedding of the binary verb predicate features.
    initializer : `InitializerApplicator`, optional (default=`InitializerApplicator()`)
        Used to initialize the model parameters.
    label_smoothing : `float`, optional (default = 0.0)
        Whether or not to use label smoothing on the labels when computing cross entropy loss.
    ignore_span_metric : `bool`, optional (default = False)
        Whether to calculate span loss, which is irrelevant when predicting BIO for Open Information Extraction.
    srl_eval_path : `str`, optional (default=`DEFAULT_SRL_EVAL_PATH`)
        The path to the srl-eval.pl script. By default, will use the srl-eval.pl included with allennlp,
        which is located at allennlp/tools/srl-eval.pl . If `None`, srl-eval.pl is not used.
    """

    def __init__(
        self,
        vocab: Vocabulary,
        text_field_embedder: TextFieldEmbedder,
        encoder: Seq2SeqEncoder,
        binary_feature_dim: int,
        embedding_dropout: float = 0.0,
        initializer: InitializerApplicator = InitializerApplicator(),
        label_smoothing: float = None,
        ignore_span_metric: bool = False,
        srl_eval_path: str = DEFAULT_SRL_EVAL_PATH,
        **kwargs,
    ) -> None:
        super().__init__(vocab, **kwargs)

        self.text_field_embedder = text_field_embedder
        self.num_classes = self.vocab.get_vocab_size("labels")

        if srl_eval_path is not None:
            # For the span based evaluation, we don't want to consider labels
            # for verb, because the verb index is provided to the model.
            self.span_metric = SrlEvalScorer(srl_eval_path, ignore_classes=["V"])
        else:
            self.span_metric = None

        self.encoder = encoder
        # There are exactly 2 binary features for the verb predicate embedding.
        self.binary_feature_embedding = Embedding(
            num_embeddings=2, embedding_dim=binary_feature_dim
        )
        self.tag_projection_layer = TimeDistributed(
            Linear(self.encoder.get_output_dim(), self.num_classes)
        )
        self.embedding_dropout = Dropout(p=embedding_dropout)
        self._label_smoothing = label_smoothing
        self.ignore_span_metric = ignore_span_metric

        check_dimensions_match(
            text_field_embedder.get_output_dim() + binary_feature_dim,
            encoder.get_input_dim(),
            "text embedding dim + verb indicator embedding dim",
            "encoder input dim",
        )
        initializer(self)

    def forward(  # type: ignore
        self,
        tokens: TextFieldTensors,
        verb_indicator: torch.LongTensor,
        tags: torch.LongTensor = None,
        metadata: List[Dict[str, Any]] = None,
    ) -> Dict[str, torch.Tensor]:

        """
        # Parameters

        tokens : TextFieldTensors, required
            The output of `TextField.as_array()`, which should typically be passed directly to a
            `TextFieldEmbedder`. This output is a dictionary mapping keys to `TokenIndexer`
            tensors.  At its most basic, using a `SingleIdTokenIndexer` this is : `{"tokens":
            Tensor(batch_size, num_tokens)}`. This dictionary will have the same keys as were used
            for the `TokenIndexers` when you created the `TextField` representing your
            sequence.  The dictionary is designed to be passed directly to a `TextFieldEmbedder`,
            which knows how to combine different word representations into a single vector per
            token in your input.
        verb_indicator: torch.LongTensor, required.
            An integer `SequenceFeatureField` representation of the position of the verb
            in the sentence. This should have shape (batch_size, num_tokens) and importantly, can be
            all zeros, in the case that the sentence has no verbal predicate.
        tags : torch.LongTensor, optional (default = None)
            A torch tensor representing the sequence of integer gold class labels
            of shape `(batch_size, num_tokens)`
        metadata : `List[Dict[str, Any]]`, optional, (default = None)
            metadata containg the original words in the sentence and the verb to compute the
            frame for, under 'words' and 'verb' keys, respectively.

        # Returns

        An output dictionary consisting of:
        logits : torch.FloatTensor
            A tensor of shape `(batch_size, num_tokens, tag_vocab_size)` representing
            unnormalised log probabilities of the tag classes.
        class_probabilities : torch.FloatTensor
            A tensor of shape `(batch_size, num_tokens, tag_vocab_size)` representing
            a distribution of the tag classes per word.
        loss : torch.FloatTensor, optional
            A scalar loss to be optimised.

        """
        embedded_text_input = self.embedding_dropout(self.text_field_embedder(tokens))
        mask = get_text_field_mask(tokens)
        embedded_verb_indicator = self.binary_feature_embedding(verb_indicator.long())
        # Concatenate the verb feature onto the embedded text. This now
        # has shape (batch_size, sequence_length, embedding_dim + binary_feature_dim).
        embedded_text_with_verb_indicator = torch.cat(
            [embedded_text_input, embedded_verb_indicator], -1
        )
        batch_size, sequence_length, _ = embedded_text_with_verb_indicator.size()

        encoded_text = self.encoder(embedded_text_with_verb_indicator, mask)

        logits = self.tag_projection_layer(encoded_text)
        reshaped_log_probs = logits.view(-1, self.num_classes)
        class_probabilities = F.softmax(reshaped_log_probs, dim=-1).view(
            [batch_size, sequence_length, self.num_classes]
        )
        output_dict = {"logits": logits, "class_probabilities": class_probabilities}
        # We need to retain the mask in the output dictionary
        # so that we can crop the sequences to remove padding
        # when we do viterbi inference in self.make_output_human_readable.
        output_dict["mask"] = mask

        if tags is not None:
            loss = sequence_cross_entropy_with_logits(
                logits, tags, mask, label_smoothing=self._label_smoothing
            )
            if not self.ignore_span_metric and self.span_metric is not None and not self.training:
                batch_verb_indices = [
                    example_metadata["verb_index"] for example_metadata in metadata
                ]
                batch_sentences = [example_metadata["words"] for example_metadata in metadata]
                # Get the BIO tags from make_output_human_readable()
                # TODO (nfliu): This is kind of a hack, consider splitting out part
                # of make_output_human_readable() to a separate function.
                batch_bio_predicted_tags = self.make_output_human_readable(output_dict).pop("tags")
                batch_conll_predicted_tags = [
                    convert_bio_tags_to_conll_format(tags) for tags in batch_bio_predicted_tags
                ]
                batch_bio_gold_tags = [
                    example_metadata["gold_tags"] for example_metadata in metadata
                ]
                batch_conll_gold_tags = [
                    convert_bio_tags_to_conll_format(tags) for tags in batch_bio_gold_tags
                ]
                self.span_metric(
                    batch_verb_indices,
                    batch_sentences,
                    batch_conll_predicted_tags,
                    batch_conll_gold_tags,
                )
            output_dict["loss"] = loss

        words, verbs = zip(*[(x["words"], x["verb"]) for x in metadata])
        if metadata is not None:
            output_dict["words"] = list(words)
            output_dict["verb"] = list(verbs)
        return output_dict

    @overrides
    def make_output_human_readable(
        self, output_dict: Dict[str, torch.Tensor]
    ) -> Dict[str, torch.Tensor]:
        """
        Does constrained viterbi decoding on class probabilities output in :func:`forward`.  The
        constraint simply specifies that the output tags must be a valid BIO sequence.  We add a
        `"tags"` key to the dictionary with the result.
        """
        all_predictions = output_dict["class_probabilities"]
        sequence_lengths = get_lengths_from_binary_sequence_mask(output_dict["mask"]).data.tolist()

        if all_predictions.dim() == 3:
            predictions_list = [
                all_predictions[i].detach().cpu() for i in range(all_predictions.size(0))
            ]
        else:
            predictions_list = [all_predictions]
        all_tags = []
        transition_matrix = self.get_viterbi_pairwise_potentials()
        start_transitions = self.get_start_transitions()
        for predictions, length in zip(predictions_list, sequence_lengths):
            max_likelihood_sequence, _ = viterbi_decode(
                predictions[:length], transition_matrix, allowed_start_transitions=start_transitions
            )
            tags = [
                self.vocab.get_token_from_index(x, namespace="labels")
                for x in max_likelihood_sequence
            ]
            all_tags.append(tags)
        output_dict["tags"] = all_tags
        return output_dict

    def get_metrics(self, reset: bool = False):
        if self.ignore_span_metric:
            # Return an empty dictionary if ignoring the
            # span metric
            return {}

        else:
            metric_dict = self.span_metric.get_metric(reset=reset)

            # This can be a lot of metrics, as there are 3 per class.
            # we only really care about the overall metrics, so we filter for them here.
            return {x: y for x, y in metric_dict.items() if "overall" in x}

    def get_viterbi_pairwise_potentials(self):
        """
        Generate a matrix of pairwise transition potentials for the BIO labels.
        The only constraint implemented here is that I-XXX labels must be preceded
        by either an identical I-XXX tag or a B-XXX tag. In order to achieve this
        constraint, pairs of labels which do not satisfy this constraint have a
        pairwise potential of -inf.

        # Returns

        transition_matrix : torch.Tensor
            A (num_labels, num_labels) matrix of pairwise potentials.
        """
        all_labels = self.vocab.get_index_to_token_vocabulary("labels")
        num_labels = len(all_labels)
        transition_matrix = torch.zeros([num_labels, num_labels])

        for i, previous_label in all_labels.items():
            for j, label in all_labels.items():
                # I labels can only be preceded by themselves or
                # their corresponding B tag.
                if i != j and label[0] == "I" and not previous_label == "B" + label[1:]:
                    transition_matrix[i, j] = float("-inf")
        return transition_matrix

    def get_start_transitions(self):
        """
        In the BIO sequence, we cannot start the sequence with an I-XXX tag.
        This transition sequence is passed to viterbi_decode to specify this constraint.

        # Returns

        start_transitions : torch.Tensor
            The pairwise potentials between a START token and
            the first token of the sequence.
        """
        all_labels = self.vocab.get_index_to_token_vocabulary("labels")
        num_labels = len(all_labels)

        start_transitions = torch.zeros(num_labels)

        for i, label in all_labels.items():
            if label[0] == "I":
                start_transitions[i] = float("-inf")

        return start_transitions

    default_predictor = "semantic-role-labeling"


def write_to_conll_eval_file(
    prediction_file: TextIO,
    gold_file: TextIO,
    verb_index: Optional[int],
    sentence: List[str],
    prediction: List[str],
    gold_labels: List[str],
):
    """
    .. deprecated:: 0.8.4
       The `write_to_conll_eval_file` function was deprecated in favor of the
       identical `write_bio_formatted_tags_to_file` in version 0.8.4.

    Prints predicate argument predictions and gold labels for a single verbal
    predicate in a sentence to two provided file references.

    The CoNLL SRL format is described in
    [the shared task data README](https://www.lsi.upc.edu/~srlconll/conll05st-release/README).

    This function expects IOB2-formatted tags, where the B- tag is used in the beginning
    of every chunk (i.e. all chunks start with the B- tag).

    # Parameters

    prediction_file : TextIO, required.
        A file reference to print predictions to.
    gold_file : TextIO, required.
        A file reference to print gold labels to.
    verb_index : Optional[int], required.
        The index of the verbal predicate in the sentence which
        the gold labels are the arguments for, or None if the sentence
        contains no verbal predicate.
    sentence : List[str], required.
        The word tokens.
    prediction : List[str], required.
        The predicted BIO labels.
    gold_labels : List[str], required.
        The gold BIO labels.
    """
    warnings.warn(
        "The 'write_to_conll_eval_file' function has been deprecated in favor of "
        "the identical 'write_bio_formatted_tags_to_file' function.",
        DeprecationWarning,
    )
    write_bio_formatted_tags_to_file(
        prediction_file, gold_file, verb_index, sentence, prediction, gold_labels
    )



from allennlp_models.syntax.constituency_parser.constituency_parser_model import (
    SpanConstituencyParser,
)
from allennlp_models.syntax.constituency_parser.constituency_parser_predictor import (
    ConstituencyParserPredictor,
)
from allennlp_models.syntax.constituency_parser.penn_tree_bank import (
    PennTreeBankConstituencySpanDatasetReader,
)



from allennlp_models.ner.crf_tagger import CrfTagger
from allennlp_models.ner.ontonotes_ner_reader import OntonotesNamedEntityRecognition
from allennlp_models.ner.conll2000_reader import Conll2000DatasetReader
from allennlp_models.ner.ccgbank_reader import CcgBankDatasetReader
from allennlp_models.ner.conll2003_reader import Conll2003DatasetReader
from allennlp_models.ner.sentence_tagger_predictor import SentenceTaggerPredictor



from allennlp_models.rc.dialog_qa.quac import QuACReader
from allennlp_models.rc.dialog_qa.dialog_qa_model import DialogQA
from allennlp_models.rc.dialog_qa.dialog_qa_predictor import DialogQAPredictor



from allennlp_models.syntax.semantic_dependencies.semantic_dependencies_reader import (
    SemanticDependenciesDatasetReader,
)
from allennlp_models.syntax.semantic_dependencies.graph_parser_model import GraphParser



from allennlp_models.sentiment.biattentive_classification_network_model import (
    BiattentiveClassificationNetwork,
)
from allennlp_models.sentiment.stanford_sentiment_tree_bank_reader import (
    StanfordSentimentTreeBankDatasetReader,
)



from allennlp_models.rc.transformer_qa.transformer_squad import TransformerSquadReader
from allennlp_models.rc.transformer_qa.transformer_qa_model import TransformerQA
from allennlp_models.rc.transformer_qa.transformer_qa_predictor import TransformerQAPredictor



from allennlp_models.syntax.biaffine_dependency_parser.biaffine_dependency_parser import (
    BiaffineDependencyParser,
)
from allennlp_models.syntax.biaffine_dependency_parser.biaffine_dependency_parser_predictor import (
    BiaffineDependencyParserPredictor,
)
from allennlp_models.syntax.biaffine_dependency_parser.universal_dependencies_reader import (
    UniversalDependenciesDatasetReader,
)
